@@ -1,21 +1,31 @@
+from .detokenizer import Detokenizer
 from .latin_sentence_tokenizer import LatinSentenceTokenizer
+from .latin_word_detokenizer import LatinWordDetokenizer
 from .latin_word_tokenizer import LatinWordTokenizer
 from .line_segment_tokenizer import LineSegmentTokenizer
 from .null_tokenizer import NullTokenizer
 from .range_tokenizer import RangeTokenizer
+from .string_detokenizer import StringDetokenizer
 from .string_tokenizer import StringTokenizer
 from .tokenizer import Tokenizer
+from .whitespace_detokenizer import WhitespaceDetokenizer
 from .whitespace_tokenizer import WhitespaceTokenizer
+from .zwsp_word_detokenizer import ZwspWordDetokenizer
 from .zwsp_word_tokenizer import ZwspWordTokenizer
 
 __all__ = [
+    "Detokenizer",
     "LatinSentenceTokenizer",
+    "LatinWordDetokenizer",
     "LatinWordTokenizer",
     "LineSegmentTokenizer",
     "NullTokenizer",
     "RangeTokenizer",
+    "StringDetokenizer",
     "StringTokenizer",
     "Tokenizer",
+    "WhitespaceDetokenizer",
     "WhitespaceTokenizer",
+    "ZwspWordDetokenizer",
     "ZwspWordTokenizer",
 ]
